%% Simulation parameters
dt = 1e-2; % time step, s
simdur = 50000; % total simulation time, s
tind = 1; % time step number for indexing
t = 0; % simulation time variable, s
x = 0; % position, m
y = 0; % position, m
z = 0;


L = 2.5;
boundLow = 0;
boundHigh = L;
% numholes = 0;
% numtimesteps = 5000;
% len = simdur;
% [x, y, z, ~, angs2, angs1] = GetRandomFlight(boundLow, boundHigh, numholes, numtimesteps, len);


%% Model parameters
% number of grid cells (thresholded saturating neurons)
NmEC = 125;
% number of place cells (inputs to grid cells)
N = 12;
N1 = N*N*N; % they say 200, but if the env is square, using a square number instead
% growth rate for onset of adaptation
b1 = 0.1;
% growth rate of inactivation of adaptation
b2 = b1/3;
% peak (saturating) value of grid cell activation
phiSat = 1;%30;
% target mean activity (dynamic normalization tries to attain this mean level)
a0 = 0.1*phiSat;
% a0 = 0.1;
% target sparsense (dynamic normalization aims for a specific sparseness too)
s0 = 0.3;
% learning rate for plasticity from place to grid cells
% epsilon = 1e-5;
epsilon = 0.002;
% place field standard deviation
sigma_p = 0.05*L; % cm
% max weight to a grid cell (not in paper?)
maxWeight = 1;%0.08;

% for weighted running average of input and postsynaptic activities:
% runningAverageWeight = 0.3;
eta = 0.05;

% animal velocity (important but not specified in paper or in the authors'
% personal communication to me--just guessing based on their Figure 1A)
% vel = 0.08;
vel = 1;
b3 = 0.01;
b4 = 0.1;



%% Model variables
W = rand(NmEC, N1);
% Normalize weights so the sum of the squares is 1
% normFactor = repmat(sum(W,2),1,N1);
for i = 1:NmEC
    W(i,:)= maxWeight*W(i,:)./norm(W(i,:));
end
mu= 0;

%% Place cell properties
cellPositions = (mod(0:(N-1),N)+0.5)*boundHigh/N;
[X, Y, Z] =ndgrid(cellPositions,cellPositions, cellPositions);
placeCenters = [reshape(X,1,[]); reshape(Y,1,[]);reshape(Z,1,[])];

%% Collateral weights
rho = 0.1;
c = 0.2;
nu = 0.8;
prefHD = linspace(0, 4*pi, NmEC);
tau = 25;
l = vel*tau*10^-2;
sigma_f = 0.2*L;
kappa = 0.05;
W1 = zeros(NmEC);
x_aux = placeCenters(1,randperm(size(placeCenters,2), NmEC));
y_aux = placeCenters(2,randperm(size(placeCenters,2), NmEC));
z_aux = placeCenters(3,randperm(size(placeCenters,2), NmEC));
for i = 1:NmEC
    for k = 1:NmEC
        wik = 2*pi + atan2(y_aux(k)-y_aux(i),x_aux(k)-x_aux(i))*...
            (1 - atan2(z_aux(k)-z_aux(i), ((x_aux(k)-x_aux(i))^2+ (y_aux(k)-y_aux(i))^2)^0.5));
        coswik = cos(wik);
        dki = (x_aux(i) - (x_aux(k)+l*coswik))^2 + (y_aux(i) - (y_aux(k)+l*coswik))^2 ...
            + (z_aux(i) - (z_aux(k)+l*coswik))^2;
        fik = (c + (1-c)*exp(nu*(cos(prefHD(i) - wik)-1)))...
            *(c + (1-c)*exp(nu*(cos(prefHD(k) - wik)-1)));
        W1(i,k) = max(fik*exp(-dki/2/(sigma_f^2)) - kappa,0);
    end
end

% No weights >1, not mentioned in paper
W1(W1>1) = 1;

% Normalize weights so the sum of the squares is 1
for ind = 1:NmEC
    if (norm(W1(ind,:))~= 0)
        W1(ind,:) = maxWeight*W1(ind,:)./norm(W1(ind,:));
    end
end
%% Initial conditionsW
% Output of transfer function:
psi = zeros(simdur/dt,NmEC);
% Activation variable:
alpha = zeros(1,NmEC);
% Inactivation variable:
beta = zeros(1,NmEC);
% Input vector:
r = zeros(1,N1);
% Input from place cells to grid cells:
h = zeros(1,NmEC);

meanpsi = psi(1,:);
meanr = r;
curDir1 = 0;
curDir2 = 0;
%% Main simulation loop
x(1) = L/2;
y(1) = L/2;
z(1) = L/2;

while t<simdur
    tind = tind+1;
    t = dt*tind;
    %     curDir(tind) = 2*pi+ (angs1(tind) - 0)*(cos(0) - cos(angs2(tind)));
    randDir = normrnd(curDir1, 0.15);
    randDir2 =  normrnd(curDir2, 0.1);
    tempx = x(tind-1)+dt*vel*cos(randDir2)*cos(randDir);
    tempy = y(tind-1)+dt*vel*cos(randDir2)*sin(randDir);
    tempz = z(tind-1) + dt*vel*sin(randDir2);
    while (tempx<boundLow || tempx >boundHigh || ...
            tempy <boundLow || tempy >boundHigh || ...
            tempz <boundLow || tempz >boundHigh)
        randDir = normrnd(curDir1, 2*pi);
        randDir2 =  normrnd(curDir2, 2*pi);
        tempx = x(tind-1)+dt*vel*cos(randDir2)*cos(randDir);
        tempy = y(tind-1)+dt*vel*cos(randDir2)*sin(randDir);
        tempz = z(tind-1) + dt*vel*sin(randDir2);
    end
    
    curDir1 = randDir; 
    curDir2 = randDir2;     
    x(tind) = tempx;
    y(tind) = tempy;
    z(tind) = tempz;
  
    % Place cell activation at this location
    squareDists = (x(tind)-placeCenters(1,:)).^2 + (y(tind)-placeCenters(2,:)).^2 ...
        + (z(tind)-placeCenters(3,:)).^2;
    
    %     r = phiSat*exp(-squareDists/2/placeSTD^2);
%     velMod = 30;
    r = phiSat*exp(-squareDists/2/(sigma_p^2));
    if(max(r) > min(r))
        r = (r-min(r))./(max(r)-min(r));
    else
        r = r/min(r);
    end
        % Update adaptation variables
    alpha = alpha + b1*(h - beta - alpha);
    beta = beta + b2*(h - beta);
    
    % Place to grid input
    if(tind>tau)
        
        ang(tind-tau) = mod((curDir1*(1- curDir2)), 4*pi);
        f = c + (1-c)*exp(nu*(cos(prefHD - ang(tind-tau))-1));
        h = f.*((W*r')' + rho*(W1*psi(tind-tau,:)')');
    else
        h = (W*r')';
    end
    % Iteratively set the threshold and gain on this time step to produce
    % the desired mean activity and sparseness
    a = -100;
    s = -100;
    g = 1;
    mu = mu + b3*(sum(psi(tind,:))/NmEC-a0);
    iter = 0;
    % None of this is mentioned in the original paper:
    % use relative tolerance of 10% (from Bailu's code)
    while abs(a-a0)>(0.1*a0) || abs(s-s0)>(0.1*s0)
        
        %         psi(tind,:) = phiSat*2/pi*atan(pi/2*g*(thresholdedActivity));
        thresholdedActivity = alpha-mu;
        thresholdedActivity(thresholdedActivity<0) = 0;
        thresholdedActivity(thresholdedActivity>0) = 1;
        psi(tind,:) = phiSat*2/pi*atan(pi/2*g*(alpha-mu)).*thresholdedActivity;
        if (max(psi(tind,:))>min(psi(tind,:)))
            psi(tind,:) = (psi(tind,:)-min(psi(tind,:)))/(max(psi(tind,:))-min(psi(tind,:)));
        else
            psi(tind,:) = (psi(tind,:)/min(psi(tind,:)));
        end
        % average activity:
        a = sum(psi(tind,:))/NmEC;
        
        % sparseness (Bailu's script floors this at 1e-6)        
        s = sum(psi(tind,:))^2/(NmEC*sum(psi(tind,:).^2));
        if (s<1e-6)
            s = 1e-6;
        end
        
        mu = mu + b3*(a-a0);
        % manuscript does not mention the squared mean here, I don't think:
%         g = g + b4*(g+0.01)*a^2*(s-s0);
        g = g + b4*g*(s-s0);
        
        % keep threshold below max activation variable
        if mu > max(alpha)
            mu = alpha - 0.1;
        end
        % limits on gain:
%         if g > 0.5
%             g = 0.5;
%         elseif g < 0.01
%             g = 0.01;
%         end
        
        % Bailu's code gives up after 50 rounds
        iter = iter+1;
        if iter>50
            break
        end
    end
    
    % Update mean input and grid cell activities for learning
    meanpsi = meanpsi*(1-eta) + eta*psi(tind,:);
    meanr = meanr*(1-eta) + eta*r;
    
    % Learning
    W = max(W + epsilon*(psi(tind,:)'*r - meanpsi'*meanr),0);
    % No weights >1, not mentioned in paper
    W(W>1) = 1;
   
    for ind = 1:NmEC
       if (norm(W(ind,:))~= 0)
            W(ind,:) = maxWeight*W(ind,:)./norm(W(ind,:));
        end
    end
end

%% Draw a few examples of weights from place cells onto grid cells
% figure;
% subplot(131)
% imagesc(reshape(W(5,:),14,14))
% axis equal
% subplot(132)
% imagesc(reshape(W(15,:),14,14))
% axis equal
% title('Weights from place cells onto 3 grid cells')
% subplot(133)
% imagesc(reshape(W(25,:),14,14))
% axis equal


scatter3(x(1000:tind),y(1000:tind), z(1000:tind), [],psi(1000:tind,10),'filled');
% save('GridData3', 'Sout', 'angs', 'posx','posy', '-v7.3');

